/**
 * ----------------------------------------------------------------------------
 * Digital Guitar Circuitry
 * This code was created by Adam Jackman (afj8@cornell.edu) and Pouria  
 * Pezeshkian (pp332@cornell.edu), parts of which are borrowed from Ruibing   
 * Wang, Bruce Land, and National Semiconductors. Use of this code is not  
 * permitted without the permission of Adam and Pouria.
 * ----------------------------------------------------------------------------
 * Created in May 2011
 **/

/**Rotary Encoder Connections**/
//Connect Volume knob rotary output to A.0
//Connect Tone1 knob rotary output to A.1
//Connect Tone2 knob rotary output to A.2
//Connect Volume knob direction output to A.3
//Connect Tone1 knob direction output to A.4
//Connect Tone2 knob direction output to A.5
//Connect NEXT button push to A.6
//Connect PREV button push to A.7
 
/**SPI Connections**/
//connect MOSI B.5 to DI
//connect MISO B.6 to DO NOT USED
//connect SCLK B.7 to CLK
//connect B.4 to chip select for R1, R2
//connect B.0 to chip select for R3, R4
//RS and SHDN are not needed here and held at 5V

/**MUX Bit Connections**/
//connect B1 to MUX LSB (A)
//connect B2 to MUX (B)
//connect B3 to MUX MSB (C)

/**LCD Connections**/
/*
 * Use an 2x16 alphanumeric LCD connected to PORTC as follows:
 *  
 *    [LCD] -    [Mega644 Pin]
 *    1 GND -    GND
 *    2 +5V -    VCC
 *    3 VLC 10k trimpot wiper (trimpot ends go to +5 and gnd) 
 *    4 RS -    PC0
 *    5 RD -    PC1
 *    6 EN -    PC2
 *    11 D4 -    PC4
 *    12 D5 -    PC5
 *    13 D6 -    PC6
 *    14 D7 -    PC7 
 */

 /**SWITCH Connections**/
 /* Connect SWITCH as shown
 *        __
 *        \ \
 * ________\_\_______
 *|                  |
 *|      SWITCH      |
 *|__________________|
 *  |  |  |  |  |  |
 * D2 D1 D0 Vcc NC NC
 *
 *
 */
 
#include <inttypes.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lcd_lib.h"

const int8_t LCD_initialize[] PROGMEM = "LCD Initialized\0";
const int8_t LCD_ton1[] PROGMEM = "TN1:\0";
const int8_t LCD_ton2[] PROGMEM = "TN2:\0";
const int8_t LCD_vol[] PROGMEM = "VOL:\0";
const int8_t LCD_preset[] PROGMEM = "PST\0";
const int8_t LCD_switch[] PROGMEM = "PK\0";

int8_t lcd_buffer[17];    // LCD display buffer
int8_t lcd_buffer1[17];    // LCD display buffer
int8_t lcd_buffer2[17];    // LCD display buffer
int8_t lcd_buffer3[17];    // LCD display buffer

//timeout value for LCD task
#define timeLCD 200 
//timeout value for debounce code
#define timeDebounce 20
#define increment 17

// running time counter for tasks
volatile unsigned char time1, time2;

// redefine brackets
#define begin {
#define end }

// bit handling macros for i/o registers
#define READ(U, N) ((U) >> (N) & 1u)
#define SET(U, N) ((void)((U) |= 1u << (N)))
#define CLR(U, N) ((void)((U) &= ~(1u << (N))))
#define FLIP(U, N) ((void)((U) ^= 1u << (N)))

//values represent the settings for each guitar preset.
volatile char tone1[8] = {255,102,0,255,255,255,255,255};
volatile char tone2[8] = {255,255,255,255,102,0,255,255};
volatile char volume[8] = {255,255,255,255,255,255,136,51};
volatile char switchpos[8] = {1,1,1,3,3,3,5,5};


//curPreset is the value of the currently selected preset
//which determines the potentiometer positions and the guitar pickup.
volatile char curPreset = 0;

//stores the previous SWITCH position for debounce code
volatile char prev_Value_SWITCH = 0;
//stores the previous ROTARY inputs for change determination on PCINT
volatile char prev_Value_ROTARIES = 0;
//stores the previous PUSHBUTTON inputs for rotary encoder buttons
volatile char prev_Value_PUSHBUTTONS = 0;

//**********************************************************
//Initialize data directions, SPI, LCD, timer0, and schedule taskLCD
int main(void)
begin
    //set up i/o data directions
	
	//D0,D1,D2 are attached to guitar switch pins
    DDRD = (1<<DDD0) | (1<<DDD1) | (1<<DDD2); 
	//**************************REMOVE ABOVE LINE??*********************//
	
	//PORTB is connected to analog MUX and digital pots, DDB.6 not used
	DDRB = (1<<DDB0) | (1<<DDB1) | (1<<DDB2) | (1<<DDB3) | (1<<DDB4) | (1<<DDB5) | (1<<DDB7) ; 
    
	//PORTA set to all inputs, senses encoders and push buttons
    DDRA = 0;
	//Enable Pull-up resistors for two push buttons on A.6 and A.7
	PORTA = 0xC0;
	
    //Enable interrupts for rotary encoder turning pins of PORTA
    PCMSK0 = 0b00000111;
	//Enable pin change interrupt for PORTA
    PCICR = (1 << PCIE0);
	
    //set up SPI
    //bit 7 SPIE=0 no ISR
    //bit 6 SPE=1 enable spi
    //bit 5 DORD=0 msb first 
    //bit 4 MSTR=1 Mega32 is spi master
    //bit 3 CPOL=1 clock polarity
    //bit 2 CPHA=1 clock phase
    //bit 1,0 rate sel=00 along with SPRC=1 sets clk to f/2 = 8 Mhz
    SPCR = (1<<SPE) | (1<<MSTR) | (1<<CPOL) | (1<<CPHA) ; //SPCR SPCR = 0b01011100 ;                                                                               
    SPSR = (1<<SPI2X) ; //SPSR = 1; 
    
	//initialize chip select bits for pots
    SET(PORTB,4) ;
    SET(PORTB,0) ;

    /*Code in section below provided by Ruibing Wang*/
    /* (rw98@cornell.edu). More by Ruibing Wang below*/
    // start the LCD 
    init_lcd();
    LCDclr();
    
    //init the task timer
    time1=timeLCD;
    time2=timeDebounce;
    
    //set up timer 0 for 1 mSec ticks
    TIMSK0 = 2;        //turn on timer 0 cmp match ISR 
    OCR0A = 249;      //set the compare reg to 250 time ticks
    TCCR0A = 0b00000010; // turn on clear-on-match
    TCCR0B = 0b00000011;    // clock prescalar to 64
    /*End of Section*/

    //Write strings to LCD, which remain permanently
    CopyStringtoLCD(LCD_preset, 0, 0);    
    CopyStringtoLCD(LCD_vol, 9, 0);
    CopyStringtoLCD(LCD_ton1, 0, 1);
    CopyStringtoLCD(LCD_ton2, 8, 1);
	CopyStringtoLCD(LCD_switch, 5, 0);
	
	//Global interrupt enable
    sei();

    //main task scheduler loop for LCD task every 200 ms
	//as low priority task outside of ISR
    while (1)
    begin 
        if (time1==0) { time1=timeLCD; taskLCD();}
    end

end

//**********************************************************
//Pin Change ISR for PORTA for rotating encoder input
ISR (PCINT0_vect)
begin
	char rotaries = PINA & 0b00000111;
    //DETERMINE WHAT PIN CHANGED
	
	//XOR two values returns a number where the 1 is the bit that changed
	char rotary_Turned = rotaries ^ prev_Value_ROTARIES;
	
	if (rotary_Turned == 0x01) //Volume Encoder Turned
    {
        if (PINA & 0b00001000)//CCW
		{
			if (volume[curPreset] >= increment)
				volume[curPreset]-=increment;
		}
		else
		{
			if (volume[curPreset] <= 255-increment)
				volume[curPreset]+=increment;
		}
	}
	if (rotary_Turned == 0x02) //Tone1 Turned
    {
        if (PINA & 0b00010000)//CCW
		{
			if (tone1[curPreset] >= increment)
				tone1[curPreset]-=increment;
		}
		else
		{
			if (tone1[curPreset] <= 255-increment)
				tone1[curPreset]+=increment;
		}
	}
	if (rotary_Turned == 0x04) //Tone2 Turned
    {
        if (PINA & 0b00100000)//CCW
		{
			if (tone1[curPreset] >= increment)
				tone2[curPreset]-=increment;
		}
		else
		{
			if (tone2[curPreset] <= 255-increment)
				tone2[curPreset]+=increment;
		}
	}
	
	prev_Value_ROTARIES = rotaries;
	//Upon changing, resistor positions are updated.
	setResistors(volume[curPreset], tone1[curPreset], 128, tone2[curPreset]);
end


//**********************************************************
//timer 0 compare ISR, for 1 ms time base and SWITCH debouncing
ISR (TIMER0_COMPA_vect)
begin      
    /*If buttons are debounced for preset selection, don't forget
	to update resistors */
	
	char val;
	char pushbuttons;
    //Update the task time
    if (time1>0)    --time1;
    if (time2>0)    --time2;
    if (time2==0)
    {
		//Debounce the SWITCH state
		val = PIND & 0b00000111;
  		if (val != prev_Value_SWITCH)
        {
				if (val == 1) //NECK set to MUX output
				{
					//000
					CLR(PORTB,1);
					CLR(PORTB,2);
					CLR(PORTB,3);
					switchpos[curPreset]=1;
				}
				if (val == 2) //MIDDLE set to MUX output
				{
					//001
					SET(PORTB,1);
					CLR(PORTB,2);
					CLR(PORTB,3);
					switchpos[curPreset]=3;
				}
				if (val == 4) //BRIDGE set to MUX output
				{
					//010
					CLR(PORTB,1);
					SET(PORTB,2);
					CLR(PORTB,3);
					switchpos[curPreset]=5;
				}
				if (val == 3)
				{
					//NULL CASE
					CLR(PORTB,1);
					CLR(PORTB,2);
					CLR(PORTB,3);
					switchpos[curPreset]=2;
				}
				if (val == 6) 
				{
					//NULL CASE
					CLR(PORTB,1);
					SET(PORTB,2);
					CLR(PORTB,3);
					switchpos[curPreset]=4;
				}
				if (val == 0) //STATE WILL NOT EXIST
				{
					//NULL CASE
					CLR(PORTB,1);
					SET(PORTB,2);
					CLR(PORTB,3);
					switchpos[curPreset]=2;
				}
				//no other states are possible
		}
		
		//Debounce the rotary encoder push buttons
		pushbuttons = ((PINA & 0b11000000) >> 6);
  		if (pushbuttons != prev_Value_PUSHBUTTONS)
        {
				if (pushbuttons == 1) //NEXT BUTTON PRESSED
				{
					if (curPreset<7)
						curPreset++;
				}
				if (pushbuttons == 2) //PREV BUTTON PRESSED
				{
					if (curPreset>0)
						curPreset--;
				}
				if (switchpos[curPreset]==1)
				{
					CLR(PORTB,1);
					CLR(PORTB,2);
					CLR(PORTB,3);
				}
				if (switchpos[curPreset]==2)
				{
					CLR(PORTB,1);
					CLR(PORTB,2);
					CLR(PORTB,3);
				}
				if (switchpos[curPreset]==3)
				{
					SET(PORTB,1);
					CLR(PORTB,2);
					CLR(PORTB,3);
				}
				if (switchpos[curPreset]==4)
				{
					CLR(PORTB,1);
					SET(PORTB,2);
					CLR(PORTB,3);
				}
				if (switchpos[curPreset]==5)
				{
					CLR(PORTB,1);
					SET(PORTB,2);
					CLR(PORTB,3);
				}
				setResistors(volume[curPreset], tone1[curPreset], 128, tone2[curPreset]);

		}
		prev_Value_PUSHBUTTONS = pushbuttons;
		prev_Value_SWITCH = val;
		time2 = timeDebounce;
    }

end  

//**********************************************************
// Sets the digital potentiometer values for both chips 
// via SPI communication from the mega644 device via PORTB.
void setResistors(char rB1, char rB2, char rB3, char rB4)
{
        //sets resistor 00 on chip 1
        CLR(PORTB,4) ;   //chip select low begins conversion 
        //send 8 bits, last two are A1 A0            
        SPDR = 0b00000000 ;
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SPDR = rB1 ; 
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SET(PORTB,4) ;  //chip select high ends conversion
        
        //sets resistor 01 on chip 1
        CLR(PORTB,4) ;
        //send 8 bits, last two are A1 A0
        SPDR = 0b00000001 ;
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SPDR = rB2 ; 
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SET(PORTB,4) ; //chip select high ends conversion

        /*//sets resistor 00 on chip 2
        CLR(PORTB,0) ;   //chip select low begins conversion 
        //send 8 bits, last two are A1 A0            
        SPDR = 0b00000000 ;
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SPDR = rB3 ; 
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SET(PORTB,0) ;  //chip select high ends conversion*/
        
        //sets resistor 01 on chip 2
        CLR(PORTB,0) ;
        //send 8 bits, last two are A1 A0
        SPDR = 0b00000001 ;
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SPDR = rB4 ; 
        //wait until SPIF is set (SPI done)
        while (!(SPSR & (1<<SPIF))) ;
        SET(PORTB,0) ; //chip select high ends conversion
}


//**********************************************************
// LCD setup - METHOD BY RUIBING WANG, MODS BY BRUCE LAND
void init_lcd(void) 
begin
    LCDinit();    //initialize the display
    LCDcursorOFF();
    LCDclr();     //clear the display
    LCDGotoXY(0,0);
    CopyStringtoLCD(LCD_initialize, 0, 0);
end    


//**********************************************************
// taskLCD writes to LCD every 200 mSec
void taskLCD (void) 
begin
	//Prints current preset value
	sprintf(lcd_buffer,"%-i",curPreset+1);
    LCDGotoXY(3, 0);
    LCDstring(lcd_buffer, strlen(lcd_buffer));    
    
	//Prints switch position (guitar pickup choice)
	sprintf(lcd_buffer,"%-i",switchpos[curPreset]);
    LCDGotoXY(7, 0);
    LCDstring(lcd_buffer, strlen(lcd_buffer));   

	//The following code prints three progress bars
	//for the preset settings. Each bar is 3 LCD 
	//characters in length, and represents the setting
	//which is a number from 0 to 256.
    LCDGotoXY(13, 0);    
	LCDprogressBar(volume[curPreset],255,3);

    LCDGotoXY(4, 1);
    LCDprogressBar(tone1[curPreset],255,3);

    LCDGotoXY(12, 1);
	LCDprogressBar(tone2[curPreset],255,3);
end

